'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var htmlparser2 = require('htmlparser2');
var terser = require('terser');
var CleanCSS = require('clean-css');
var RelateUrl = require('relateurl');

function _interopDefaultLegacy (e) { return e && typeof e === 'object' && 'default' in e ? e : { 'default': e }; }

var CleanCSS__default = /*#__PURE__*/_interopDefaultLegacy(CleanCSS);
var RelateUrl__default = /*#__PURE__*/_interopDefaultLegacy(RelateUrl);

const noop = () => { };
const toLowerCase = (value = '') => value.toLowerCase();

const defaultOptions = {
  log: noop,
  name: toLowerCase,

  decodeEntities: false,
  caseSensitive: false,

  removeTagWhitespace: false,
  quoteCharacter: null,

  removeAttributeQuotes: false,
  collapseBooleanAttributes: false,
  removeEmptyAttributes: false,
  preventAttributesEscaping: false,
  removeRedundantAttributes: false,
  removeScriptTypeAttributes: false,
  removeStyleLinkTypeAttributes: false,

  customEventAttributes: null,

  removeComments: false,

  minifyJS: null,
  minifyCSS: null,
  minifyURLs: null,

  keepClosingSlash: false,
  useShortDoctype: false,
  customAttrCollapse: null,

  // Below are not implemented or partially completed options
  sortAttributes: false,
  sortClassName: false,
  processScripts: null,

  ignoreCustomComments: [
    /^!/,
    /^\s*#/
  ],

  ignoreCustomFragments: [
    /<%[\s\S]*?%>/,
    /<\?[\s\S]*?\?>/
  ],

  // whitespace
  collapseWhitespace: false,
  collapseInlineTagWhitespace: true,
  conservativeCollapse: false,
  preserveLineBreaks: false,
  trimCustomFragments: false,

  removeEmptyElements: false,
  includeAutoGeneratedTags: true,
  html5: true // not sure how to handle this
};

class Context extends Map {
  extend(context = new Map()) {
    context.forEach((value, key) => {
      this.set(key, value);
    });
  }
}

class DefaultContext extends Map {
  set() {
    throw new Error('Default context doens`t have set method.');
  }
}
const defaultContext = new DefaultContext();

const uniqueId = (value) => {
  let id;

  do {
    id = Math.random().toString(36).replace(/^0\.[0-9]*/, '');
  } while (~value.indexOf(id));

  return id;
};

const EMPTY = '';
const SINGLE_SPACE = ' ';

const collapseWhitespaceAll = (str = EMPTY) => {
  // Non-breaking space is specifically handled inside the replacer function here:
  return str.replace(/[ \n\r\t\f\xA0]+/g, function (spaces) {
    return spaces === '\t' ? '\t' : spaces.replace(/(^|\xA0+)[^\xA0]+/g, '$1 ');
  });
};

const trimWhitespace = (str = EMPTY) => {
  return str.replace(/^[ \n\r\t\f]+/, EMPTY)
    .replace(/[ \n\r\t\f]+$/, EMPTY);
};

const trimStart = (str = EMPTY, hasLinebreak, conservativeCollapse = false) => {
  return str.replace(/^[ \n\r\t\f\xA0]+/, function (spaces) {
    const conservative = !hasLinebreak && conservativeCollapse;
    if (conservative && spaces === '\t') {
      return '\t';
    }
    return spaces
      .replace(/^[^\xA0]+/, EMPTY)
      .replace(/(\xA0+)[^\xA0]+/g, '$1 ') || (conservative ? SINGLE_SPACE : EMPTY);
  });
};

const trimEnd = (str = EMPTY, hasLinebreak, conservativeCollapse = false) => {
  return str.replace(/[ \n\r\t\f\xA0]+$/, function (spaces) {
    const conservative = !hasLinebreak && conservativeCollapse;
    if (conservative && spaces === '\t') {
      return '\t';
    }
    return spaces
      .replace(/[^\xA0]+(\xA0+)/g, ' $1')
      .replace(/[^\xA0]+$/, EMPTY) || (conservative ? SINGLE_SPACE : EMPTY);
  });
};

const defaultsCollapseOptions = {
  preserveLineBreaks: false,
  conservativeCollapse: false
};

const collapseWhitespace = (str = EMPTY, opts = {}, trimLeft = false, trimRight = false, collapseAll = false) => {
  let text = str;
  const options = { ...defaultsCollapseOptions, ...opts };

  let lineBreakBefore = '';
  let lineBreakAfter = '';

  if (options.preserveLineBreaks) {
    text = text.replace(/^[ \n\r\t\f]*?[\n\r][ \n\r\t\f]*/, function () {
      lineBreakBefore = '\n';
      return '';
    }).replace(/[ \n\r\t\f]*?[\n\r][ \n\r\t\f]*$/, function () {
      lineBreakAfter = '\n';
      return '';
    });
  }

  if (trimLeft) {
    // Non-breaking space is specifically handled inside the replacer function here:
    text = trimStart(text, lineBreakBefore, options.conservativeCollapse);
  }

  if (trimRight) {
    // Non-breaking space is specifically handled inside the replacer function here:
    text = trimEnd(text, lineBreakAfter, options.conservativeCollapse);
  }

  if (collapseAll) {
    // strip non space whitespace then compress spaces to one
    text = collapseWhitespaceAll(text);
  }

  return lineBreakBefore + text + lineBreakAfter;
};

const preProcessInput = (input = EMPTY, options = defaultOptions, ctx = defaultContext) => {
  let html = input;

  if (options.collapseWhitespace) {
    html = collapseWhitespace(html, options, true, true);
  }

  let uidAttr = null;
  let uidPattern = null;
  const ignoredCustomMarkupChunks = [];
  const uidIgnore = null;

  const customFragments = options.ignoreCustomFragments.map(re => re.source);

  if (customFragments.length) {
    const reCustomIgnore = new RegExp('\\s*(?:' + customFragments.join('|') + ')+\\s*', 'g');

    // temporarily replace custom ignored fragments with unique attributes
    html = input.replace(reCustomIgnore, (match) => {
      if (!uidAttr) {
        uidAttr = uniqueId(input);
        uidPattern = new RegExp('(\\s*)' + uidAttr + '([0-9]+)' + uidAttr + '(\\s*)', 'g');
      }

      const token = uidAttr + ignoredCustomMarkupChunks.length + uidAttr;
      ignoredCustomMarkupChunks.push(/^(\s*)[\s\S]*?(\s*)$/.exec(match));
      return '\t' + token + '\t';
    });
  }

  ctx.set('uidAttr', uidAttr);
  ctx.set('uidPattern', uidPattern);
  ctx.set('uidIgnore', uidIgnore);
  ctx.set('ignoredCustomMarkupChunks', ignoredCustomMarkupChunks);

  const restore = (output = '') => {
    if (!uidPattern) {
      return output;
    }

    return output.replace(uidPattern, function (match, prefix, index, suffix) {
      let chunk = ignoredCustomMarkupChunks[+index][0];
      if (options.collapseWhitespace) {
        if (prefix !== '\t') {
          chunk = prefix + chunk;
        }
        if (suffix !== '\t') {
          chunk += suffix;
        }

        const trimLeft = /^[ \n\r\t\f]/.test(chunk);
        const trimRight = /[ \n\r\t\f]$/.test(chunk);

        const whitespaceOptions = {
          preserveLineBreaks: options.preserveLineBreaks,
          conservativeCollapse: !options.trimCustomFragments
        };

        return collapseWhitespace(chunk, whitespaceOptions, trimLeft, trimRight);
      }
      return chunk;
    });
  };

  return {
    html,
    restore
  };
};

class NodeWithChildren {
  constructor(type, children = []) {
    this.type = type;
    this.children = children;
  }
}
class RootDocument extends NodeWithChildren {
  constructor(children) {
    super(htmlparser2.ElementType.Root, children);
  }
}

class Element extends NodeWithChildren {
  constructor(name, attrs, children, type) {
    super(type, children);
    this.name = name;
    this.attrs = attrs;
    this.startImplied = false;
    this.endImplied = false;
  }
}

class DataNode {
  constructor(type, data) {
    this.type = type;
    this.data = data;
  }
}

class Text extends DataNode {
  constructor(data) {
    super(htmlparser2.ElementType.Text, data);
  }
}

class Comment extends DataNode {
  constructor(data) {
    super(htmlparser2.ElementType.Comment, data);
  }
}

class Directive extends DataNode {
  constructor(name, data) {
    super(htmlparser2.ElementType.Directive, data);
    this.name = name;
  }
}

const defaultOpts = {};

class DomHandler {
  constructor(options = defaultOpts) {
    this.options = { ...defaultOpts, ...options };

    this.setDefaults();
  }

  onparserinit(parser) {
    this.parser = parser;
  }

  // Resets the handler back to starting state
  setDefaults() {
    this.dom = [];
    this.root = new RootDocument(this.dom);
    this.tagStack = [this.root];
    this.lastNode = null;
    this.parser = null;
    this.attrs = null;
  }

  onreset() {
    this.setDefaults();
  }

  onend() {
    this.parser = null;
  }

  onclosetag(_, isImplied) {
    this.lastNode = null;

    const elem = this.tagStack.pop();
    elem.endIndex = this.parser.endIndex;
    elem.endImplied = isImplied;

    this.attrs = null;
  }

  onattribute(name, value, quote) {
    if (!this.attrs) {
      this.attrs = [];
    }

    this.attrs.push({ name, value, quote });
  }

  onopentag(name, _, isImplied) {
    const type = htmlparser2.ElementType.Tag;

    const element = new Element(name, this.attrs, undefined, type);
    element.startImplied = isImplied;

    this.addNode(element);
    this.tagStack.push(element);
    this.attrs = null;
  }

  ontext(data) {
    const { lastNode } = this;

    if (lastNode && lastNode.type === htmlparser2.ElementType.Text) {
      lastNode.data += data;
    } else {
      const node = new Text(data);
      this.addNode(node);
      this.lastNode = node;
    }
  }

  oncomment(data) {
    if (this.lastNode?.type === htmlparser2.ElementType.Comment) {
      this.lastNode.data += data;
      return;
    }

    const node = new Comment(data);
    this.addNode(node);
    this.lastNode = node;
  }

  oncommentend() {
    this.lastNode = null;
  }

  onprocessinginstruction(name, data) {
    const node = new Directive(name, data);
    this.addNode(node);
  }

  addNode(node) {
    const parent = this.tagStack[this.tagStack.length - 1];
    const previousSibling = parent.children[parent.children.length - 1];

    node.startIndex = this.parser.startIndex;
    node.endIndex = this.parser.endIndex;

    parent.children.push(node);

    if (previousSibling) {
      node.prev = previousSibling;
      previousSibling.next = node;
    }

    node.parent = parent;
    this.lastNode = null;
  }
}

const isIgnoredComment = (text = '', ignoreCustomComments = defaultOptions.ignoreCustomComments) => {
  return ignoreCustomComments.some(ignoredCommentRegex => ignoredCommentRegex.test(text));
};

const isConditionalComment = (text) => {
  return /^\[if\s[^\]]+]|\[endif]$/.test(text);
};

const removeComments$1 = (node, options = defaultOptions) => {
  const isIgnored = isIgnoredComment(node.data, options.ignoreCustomComments);

  if (isIgnored) {
    return;
  }

  const isConditional = isConditionalComment(node.data);

  if (isConditional) {
    return;
  }

  htmlparser2.DomUtils.removeElement(node);
};

const removeComments = (text = '') => {
  const start = text.match(/^\s*<!--.*/);
  return start ? text.slice(start[0].length).replace(/\n\s*-->\s*$/, '') : text;
};

const removeTrailingSemi = (text = '') => {
  return text.replace(/;$/, '');
};

const minifyJS = async (text = '', inline = false, options = defaultOptions, ctx = defaultContext) => {
  if (!options.minifyJS) {
    return text;
  }

  if (typeof options.minifyJS === 'function') {
    const code = await options.minifyJS(text, inline);
    return code;
  }

  const minifierOptions = {};

  if (typeof options.minifyJS === 'object') {
    Object.assign(minifierOptions, options.minifyJS);
  }

  minifierOptions.parse = {
    ...minifierOptions.parse,
    bare_returns: inline
  };

  let code = removeComments(text);

  const uidPattern = ctx.get('uidPattern');

  if (uidPattern) {
    const ignoredCustomMarkupChunks = ctx.get('ignoredCustomMarkupChunks');
    const uidAttr = ctx.get('uidAttr');

    code = text.replace(uidPattern, function (match, prefix, index) {
      const chunks = ignoredCustomMarkupChunks[+index];
      return chunks[1] + uidAttr + index + uidAttr + chunks[2];
    });
  }

  try {
    const result = await terser.minify(code, minifierOptions);
    code = removeTrailingSemi(result.code);
  } catch (err) {
    code = text;
    options.log(err);
  }

  return code;
};

const minifyURLs = (text, options = defaultOptions) => {
  if (!options.minifyURLs) {
    return text;
  }

  if (typeof options.minifyURLs === 'function') {
    return options.minifyURLs(text);
  }

  const relateURLOptions = {};

  if (typeof options.minifyURLs === 'string') {
    relateURLOptions.site = options.minifyURLs;
  }

  if (typeof options.minifyURLs === 'object') {
    Object.assign(relateURLOptions, options.minifyURLs);
  }

  try {
    return RelateUrl__default["default"].relate(text, relateURLOptions);
  } catch (err) {
    options.log(err);
    return text;
  }
};

// Wrap CSS declarations for CleanCSS > 3.x
// See https://github.com/jakubpawlowicz/clean-css/issues/418
const wrapCSS = (text, type) => {
  switch (type) {
    case 'inline':
      return '*{' + text + '}';
    case 'media':
      return '@media ' + text + '{a{top:0}}';
    default:
      return text;
  }
};

const unwrapCSS = (text, type) => {
  let matches;

  switch (type) {
    case 'inline':
      matches = text.match(/^\*\{([\s\S]*)\}$/);
      break;
    case 'media':
      matches = text.match(/^@media ([\s\S]*?)\s*{[\s\S]*}$/);
      break;
  }

  return matches ? matches[1] : text;
};

const minifyCSS = async (text = '', type = '', options = defaultOptions) => {
  if (!options.minifyCSS) {
    return text;
  }

  if (typeof options.minifyCSS === 'function') {
    const code = options.minifyCSS(text, type);
    return code;
  }

  const minifierOptions = {
    returnPromise: true
  };

  if (typeof options.minifyCSS === 'object') {
    Object.assign(minifierOptions, options.minifyCSS);
  }

  const urlMinifiedText = text.replace(/(url\s*\(\s*)("|'|)(.*?)\2(\s*\))/ig, function (match, prefix, quote, url, suffix) {
    return prefix + quote + minifyURLs(url, options) + quote + suffix;
  });

  let code = wrapCSS(urlMinifiedText, type);
  const cleancss = new CleanCSS__default["default"](minifierOptions);

  try {
    const result = await cleancss.minify(code);
    code = result.styles;
  } catch (errors) {
    errors.forEach(options.log);
  }

  return unwrapCSS(code, type);
};

// empty elements
const voidTags = new Set(['area', 'base', 'basefont', 'br', 'col', 'embed', 'frame', 'hr', 'img', 'input', 'isindex', 'keygen', 'link', 'meta', 'param', 'source', 'track', 'wbr']);

// non-empty tags that will maintain whitespace around them
const inlineTags = new Set(['a', 'abbr', 'acronym', 'b', 'bdi', 'bdo', 'big', 'button', 'cite', 'code', 'del', 'dfn', 'em', 'font', 'i', 'ins', 'kbd', 'label', 'mark', 'math', 'nobr', 'object', 'q', 'rp', 'rt', 'rtc', 'ruby', 's', 'samp', 'select', 'small', 'span', 'strike', 'strong', 'sub', 'sup', 'svg', 'textarea', 'time', 'tt', 'u', 'var']);

// self-closing tags that will maintain whitespace around them
const selfClosingTags = new Set(['comment', 'img', 'input', 'wbr']);

// boolean attributes
const booleanAttributes = new Set(['allowfullscreen', 'async', 'autofocus', 'autoplay', 'checked', 'compact', 'controls', 'declare', 'default', 'defaultchecked', 'defaultmuted', 'defaultselected', 'defer', 'disabled', 'enabled', 'formnovalidate', 'hidden', 'indeterminate', 'inert', 'ismap', 'itemscope', 'loop', 'multiple', 'muted', 'nohref', 'noresize', 'noshade', 'novalidate', 'nowrap', 'open', 'pauseonexit', 'readonly', 'required', 'reversed', 'scoped', 'seamless', 'selected', 'sortable', 'truespeed', 'typemustmatch', 'visible']);

const srcsetTags = new Set(['img', 'source']);

const specialContentTags = new Set(['script', 'style']);

const hasAttrName = (name, attrs) => attrs?.some(attr => attr.name === name) || false;

function canRemoveElement(tag, attrs) {
  switch (tag) {
    case 'textarea':
      return false;

    case 'audio':
    case 'script':
    case 'video':
      if (hasAttrName('src', attrs)) {
        return false;
      }
      break;

    case 'iframe':
      if (hasAttrName('src', attrs) || hasAttrName('srcdoc', attrs)) {
        return false;
      }
      break;

    case 'object':
      if (hasAttrName('data', attrs)) {
        return false;
      }
      break;

    case 'applet':
      if (hasAttrName('code', attrs)) {
        return false;
      }
      break;
  }

  return true;
}

const removeEmtpyElements = (node, options = defaultOptions) => {
  let removeElement = false;

  if (node.children.length) {
    // the node only has comment as its child remove it
    const hasOnlyCommentChildren = node.children.every(child => child.type === htmlparser2.ElementType.Comment);
    if (hasOnlyCommentChildren) {
      const commentNode = node.children.find(child => child.type === htmlparser2.ElementType.Comment);
      htmlparser2.DomUtils.removeElement(commentNode);
    }
  }

  if (
    !node.children.length &&
    !voidTags.has(node.name) &&
    canRemoveElement(node.name, node.attrs)
  ) {
    removeElement = true;
  }

  if (removeElement) {
    htmlparser2.DomUtils.removeElement(node);
  }

  return removeElement;
};

const SINGLE_QUOTE = '\'';
const DOUBLE_QUOTE = '"';
const DEFAULT_QUOTE_CHARACTER = DOUBLE_QUOTE;

const quoteCharacter = (node, options = defaultOptions) => {
  const { attrs } = node;

  if (!attrs) {
    return;
  }

  node.attrs = attrs.map(attr => {
    if (options.preventAttributesEscaping) {
      if (attr.quote === null) {
        return {
          ...attr,
          quote: ''
        };
      }

      return attr;
    }

    if (typeof attr.quote === 'undefined') {
      return attr;
    }

    let value = attr.value;
    let quote = attr.quote ?? DEFAULT_QUOTE_CHARACTER;

    if (options.quoteCharacter) {
      quote = options.quoteCharacter === SINGLE_QUOTE
        ? SINGLE_QUOTE
        : DOUBLE_QUOTE;
    } else {
      if (attr.quote) {
        const apos = (attr.value.match(/'/g) || []).length;
        const quot = (attr.value.match(/"/g) || []).length;
        quote = apos < quot ? '\'' : '"';
      }
    }

    if (quote === DOUBLE_QUOTE) {
      value = value.replace(/"/g, '&#34;');
    }

    if (quote === SINGLE_QUOTE) {
      value = value.replace(/'/g, '&#39;');
    }

    return {
      ...attr,
      value,
      quote
    };
  });
};

const useShortDoctype = (node, options = defaultOptions) => {
  if (options.useShortDoctype) {
    node.data = '!doctype' + (options.removeTagWhitespace ? '' : ' ') + 'html';
  } else {
    node.data = collapseWhitespaceAll(node.data);
  }
};

const attributesInclude = (attrs = [], name) => attrs.some(attr => attr.name === name);

const isAttributeRedundant = (tag, name, value, attrs) => {
  const attrValue = value ? trimWhitespace(value.toLowerCase()) : '';

  return (
    (tag === 'script' && name === 'language' && attrValue === 'javascript') ||
    (tag === 'form' && name === 'method' && attrValue === 'get') ||
    (tag === 'input' && name === 'type' && attrValue === 'text') ||
    (tag === 'script' && name === 'charset' && !attributesInclude(attrs, 'src')) ||
    (tag === 'a' && name === 'name' && attributesInclude(attrs, 'id')) ||
    (tag === 'area' && name === 'shape' && attrValue === 'rect')
  );
};

const removeRedundantAttributes = (attrs, node) => {
  return attrs.filter(attr => !isAttributeRedundant(node.name, attr.name, attr.value, attrs));
};

const isBooleanValue = (value) => ['true', 'false'].includes(value);

const isBooleanAttribute = (name, value) => {
  return booleanAttributes.has(name) || (name === 'draggable' && !isBooleanValue(value));
};

const collapseBooleanAttributes = (attrs) => {
  return attrs.map(attr => {
    if (isBooleanAttribute(attr.name.toLowerCase(), attr.value)) {
      return {
        ...attr,
        value: '',
        quote: undefined
      };
    }

    return attr;
  });
};

const removeAttributeQuotes = (attrs) => {
  return attrs.map(attr => {
    if (attr.value.split(' ').length > 1) {
      return attr;
    }

    return {
      ...attr,
      quote: ''
    };
  });
};

// https://mathiasbynens.be/demo/javascript-mime-type
// https://developer.mozilla.org/en/docs/Web/HTML/Element/script#attr-type
const executableScriptsMimetypes = new Set([
  'text/javascript',
  'text/ecmascript',
  'text/jscript',
  'application/javascript',
  'application/x-javascript',
  'application/ecmascript',
  'module'
]);

const keepScriptsMimetypes = new Set(['module']);

const isScriptTypeAttribute = (value) => {
  const attrValue = trimWhitespace(value.split(/;/, 2)[0]).toLowerCase();
  return attrValue === '' || executableScriptsMimetypes.has(attrValue);
};

const keepScriptTypeAttribute = (value) => {
  const attrValue = trimWhitespace(value.split(/;/, 2)[0]).toLowerCase();
  return keepScriptsMimetypes.has(attrValue);
};

const isExecutableScript = (tag, attrs) => {
  if (tag !== 'script') {
    return false;
  }

  if (!attrs) {
    return true;
  }

  return attrs.some(attr => {
    const name = attr.name.toLowerCase();

    if (name === 'type') {
      return isScriptTypeAttribute(attr.value);
    }

    return true;
  });
};

const removeScriptTypeAttributes$1 = (attrs = []) => {
  return attrs.filter(attr => {
    const { name, value } = attr;

    if (name !== 'type') {
      return true;
    }

    return !(isScriptTypeAttribute(value) && !keepScriptTypeAttribute(value));
  });
};

const isStyleLinkTypeAttribute$1 = (value) => {
  const attrValue = trimWhitespace(value).toLowerCase();
  return attrValue === '' || attrValue === 'text/css';
};

const removeScriptTypeAttributes = (attrs = []) => {
  return attrs.filter(attr => {
    const { name, value } = attr;

    if (name !== 'type') {
      return true;
    }

    return !isStyleLinkTypeAttribute$1(value);
  });
};

const EMPTY_ATTRIBUTE_REGEX = new RegExp(
  '^(?:class|id|style|title|lang|dir|on(?:focus|blur|change|click|dblclick|mouse(' +
  '?:down|up|over|move|out)|key(?:press|down|up)))$'
);

function canDeleteEmptyAttribute(tag, name, value, options) {
  const isValueEmpty = !value || /^\s*$/.test(value);
  if (!isValueEmpty) {
    return false;
  }

  return (tag === 'input' && name === 'value') || EMPTY_ATTRIBUTE_REGEX.test(name);
}

const removeEmptyAttributes = (attrs, node, options = defaultOptions) => {
  const tag = node.name;

  return attrs.filter(attr => {
    const { name, value } = attr;

    if (typeof options.removeEmptyAttributes === 'function') {
      return !options.removeEmptyAttributes(name, tag);
    }

    if (canDeleteEmptyAttribute(tag, name, value)) {
      return false;
    }

    return true;
  });
};

/* eslint-disable brace-style */

const isEventAttribute = (attrName, options = defaultOptions) => {
  const patterns = options.customEventAttributes;

  if (patterns) {
    for (let i = patterns.length; i--;) {
      if (patterns[i].test(attrName)) {
        return true;
      }
    }
    return false;
  }

  return /^on[a-z]{3,}$/.test(attrName);
};

const isUriTypeAttribute = (attrName, tag) => {
  return (
    (/^(?:a|area|link|base)$/.test(tag) && attrName === 'href') ||
    (tag === 'img' && /^(?:src|longdesc|usemap)$/.test(attrName)) ||
    (tag === 'object' && /^(?:classid|codebase|data|usemap)$/.test(attrName)) ||
    (tag === 'q' && attrName === 'cite') ||
    (tag === 'blockquote' && attrName === 'cite') ||
    ((tag === 'ins' || tag === 'del') && attrName === 'cite') ||
    (tag === 'form' && attrName === 'action') ||
    (tag === 'input' && (attrName === 'src' || attrName === 'usemap')) ||
    (tag === 'head' && attrName === 'profile') ||
    (tag === 'script' && (attrName === 'src' || attrName === 'for'))
  );
};

const isLinkType = (tag, attrs, value) => {
  if (tag !== 'link') {
    return false;
  }

  return attrs.some(attr => attr.name === 'rel' && attr.value === value);
};

const isNumberTypeAttribute = (attrName, tag) => {
  return (
    (/^(?:a|area|object|button)$/.test(tag) && attrName === 'tabindex') ||
    (tag === 'input' && (attrName === 'maxlength' || attrName === 'tabindex')) ||
    (tag === 'select' && (attrName === 'size' || attrName === 'tabindex')) ||
    (tag === 'textarea' && /^(?:rows|cols|tabindex)$/.test(attrName)) ||
    (tag === 'colgroup' && attrName === 'span') ||
    (tag === 'col' && attrName === 'span') ||
    ((tag === 'th' || tag === 'td') && (attrName === 'rowspan' || attrName === 'colspan'))
  );
};

const isSrcset = (attrName, tag) => {
  return attrName === 'srcset' && srcsetTags.has(tag);
};

function isMetaViewport(tag, attrs) {
  if (tag !== 'meta') {
    return false;
  }
  return attrs.some(attr => attr.name === 'name' && attr.value === 'viewport');
}

const isContentSecurityPolicy = (tag, attrs) => {
  if (tag !== 'meta') {
    return false;
  }

  return attrs.some(attr => {
    const { name, value } = attr;
    return name.toLowerCase() === 'http-equiv' && value.toLowerCase() === 'content-security-policy';
  });
};

const isStyleLinkTypeAttribute = (attrValue) => {
  const value = trimWhitespace(attrValue).toLowerCase();
  return (value === '' || value === 'text/css');
};

const isStyleSheet = (tag, attrs) => {
  if (tag !== 'style') {
    return false;
  }

  for (let i = 0; i < attrs.length; i++) {
    const attrName = attrs[i].name.toLowerCase();
    if (attrName === 'type') {
      return isStyleLinkTypeAttribute(attrs[i].value);
    }
  }

  return true;
};

const isMediaQuery = (tag, attrs, attrName) => {
  return (
    attrName === 'media' &&
    (isLinkType(tag, attrs, 'stylesheet') || isStyleSheet(tag, attrs))
  );
};

const cleanAttributeValue = async (attrs, node, options = defaultOptions, ctx = defaultContext) => {
  const tag = node.name;

  const promises = attrs.map(async (attr) => {
    const { name, value } = attr;

    let attrValue = value;

    if (isEventAttribute(name, options)) {
      attrValue = trimWhitespace(attrValue).replace(/^javascript:\s*/i, '');
      attrValue = await minifyJS(attrValue, true, options, ctx);
    }

    else if (name === 'class') {
      attrValue = trimWhitespace(attrValue);
      // TODO: Implement sort attributes here if not implemented later
      attrValue = collapseWhitespaceAll(attrValue);
    }

    else if (isUriTypeAttribute(name, tag)) {
      attrValue = trimWhitespace(attrValue);
      attrValue = isLinkType(tag, attrs, 'canonical')
        ? attrValue
        : minifyURLs(attrValue, options);
    }

    else if (isNumberTypeAttribute(name, tag)) {
      attrValue = trimWhitespace(attrValue);
    }

    else if (name === htmlparser2.ElementType.Style) {
      attrValue = trimWhitespace(attrValue);
      if (attrValue) {
        if (/;$/.test(attrValue) && !/&#?[0-9a-zA-Z]+;$/.test(attrValue)) {
          attrValue = attrValue.replace(/\s*;$/, ';');
        }

        attrValue = await minifyCSS(attrValue, 'inline', options);
      }
    }

    else if (isSrcset(name, tag)) {
      // https://html.spec.whatwg.org/multipage/embedded-content.html#attr-img-srcset
      attrValue = trimWhitespace(attrValue).split(/\s+,\s*|\s*,\s+/).map(function (candidate) {
        let url = candidate;
        let descriptor = '';
        const match = candidate.match(/\s+([1-9][0-9]*w|[0-9]+(?:\.[0-9]+)?x)$/);
        if (match) {
          url = url.slice(0, -match[0].length);
          const num = +match[1].slice(0, -1);
          const suffix = match[1].slice(-1);
          if (num !== 1 || suffix !== 'x') {
            descriptor = ' ' + num + suffix;
          }
        }
        return minifyURLs(url, options) + descriptor;
      }).join(', ');
    }

    else if (isMetaViewport(tag, attrs) && name === 'content') {
      attrValue = attrValue.replace(/\s+/g, '').replace(/[0-9]+\.[0-9]+/g, function (numString) {
        // "0.90000" -> "0.9"
        // "1.0" -> "1"
        // "1.0001" -> "1.0001" (unchanged)
        return (+numString).toString();
      });
    }

    else if (isContentSecurityPolicy(tag, attrs) && name.toLowerCase() === 'content') {
      attrValue = collapseWhitespaceAll(attrValue);
    }

    else if (options.customAttrCollapse && options.customAttrCollapse.test(name)) {
      attrValue = trimWhitespace(attrValue
        .replace(/ ?[\n\r]+ ?/g, '')
        .replace(/\s{2,}/g, options.conservativeCollapse ? SINGLE_SPACE : EMPTY
        )
      );
    }

    else if (tag === htmlparser2.ElementType.Script && name === 'type') {
      attrValue = trimWhitespace(attrValue.replace(/\s*;\s*/g, ';'));
    }

    else if (isMediaQuery(tag, attrs, name)) {
      attrValue = trimWhitespace(attrValue);
      attrValue = await minifyCSS(attrValue, 'media', options);
    }

    return {
      ...attr,
      value: attrValue
    };
  });

  return Promise.all(promises);
};

const normalizeAttributes = async (node, options = defaultOptions, ctx = defaultContext) => {
  const { attrs } = node;

  if (!attrs) {
    return;
  }

  let normalizedAttrs = attrs;

  if (options.removeRedundantAttributes) {
    normalizedAttrs = removeRedundantAttributes(normalizedAttrs, node);
  }

  if (options.removeScriptTypeAttributes && node.name === htmlparser2.ElementType.Script) {
    normalizedAttrs = removeScriptTypeAttributes$1(normalizedAttrs);
  }

  if (options.removeStyleLinkTypeAttributes && (node.name === htmlparser2.ElementType.Style || node.name === 'link')) {
    normalizedAttrs = removeScriptTypeAttributes(normalizedAttrs);
  }

  if (options.collapseBooleanAttributes) {
    normalizedAttrs = collapseBooleanAttributes(normalizedAttrs);
  }

  if (attrs.length) {
    normalizedAttrs = await cleanAttributeValue(normalizedAttrs, node, options, ctx);
  }

  if (options.removeAttributeQuotes) {
    normalizedAttrs = removeAttributeQuotes(normalizedAttrs);
  }

  if (options.removeEmptyAttributes) {
    normalizedAttrs = removeEmptyAttributes(normalizedAttrs, node, options);
  }

  node.attrs = normalizedAttrs;
};

// htmlparser2 already supports decoding via decodeEntities option
// this is to fix
// https://github.com/kangax/html-minifier/issues/964
const decodeEntities = (str = '', node, options = defaultOptions) => {
  let text = str;

  if (options.decodeEntities && text && !specialContentTags.has(node.parent?.name)) {
    // Escape any `&` symbols that start either:
    // 1) a legacy named character reference (i.e. one that doesn't end with `;`)
    // 2) or any other character reference (i.e. one that does end with `;`)
    // Note that `&` can be escaped as `&amp`, without the semi-colon.
    // https://mathiasbynens.be/notes/ambiguous-ampersands
    text = text
      .replace(/&((?:Iacute|aacute|uacute|plusmn|Otilde|otilde|agrave|Agrave|Yacute|yacute|Oslash|oslash|atilde|Atilde|brvbar|ccedil|Ccedil|Ograve|curren|divide|eacute|Eacute|ograve|Oacute|egrave|Egrave|Ugrave|frac12|frac14|frac34|ugrave|oacute|iacute|Ntilde|ntilde|Uacute|middot|igrave|Igrave|iquest|Aacute|cedil|laquo|micro|iexcl|Icirc|icirc|acirc|Ucirc|Ecirc|ocirc|Ocirc|ecirc|ucirc|Aring|aring|AElig|aelig|acute|pound|raquo|Acirc|times|THORN|szlig|thorn|COPY|auml|ordf|ordm|Uuml|macr|uuml|Auml|ouml|Ouml|para|nbsp|euml|quot|QUOT|Euml|yuml|cent|sect|copy|sup1|sup2|sup3|iuml|Iuml|ETH|shy|reg|not|yen|amp|AMP|REG|uml|eth|deg|gt|GT|LT|lt)(?!;)|(?:#?[0-9a-zA-Z]+;))/g, '&amp$1')
      .replace(/</g, '&lt;');
  }

  return text;
};

class Minifier {
  constructor(options = defaultOptions, ctx = defaultContext) {
    this.options = { ...defaultOptions, ...options };
    this.ctx = ctx;
  }

  async processText(node) {
    if (this.options.decodeEntities) {
      node.data = decodeEntities(node.data, node, this.options);
    }
  }

  async processTag(node) {
    // remove elements that are empty
    if (this.options.removeEmptyElements) {
      const removed = removeEmtpyElements(node, this.options);
      if (removed) {
        return;
      }
    }

    await normalizeAttributes(node, this.options, this.ctx);

    // apply preferred quote character
    quoteCharacter(node, this.options);

    await this.minify(node.children);

    switch (node.name) {
      case htmlparser2.ElementType.Script: {
        const textNode = node.children[0];
        if (this.options.minifyJS && textNode && isExecutableScript(node.name, node.attrs)) {
          textNode.data = await minifyJS(textNode.data, false, this.options, this.ctx);
        }
        break;
      }

      case htmlparser2.ElementType.Style: {
        const textNode = node.children[0];
        if (this.options.minifyCSS && textNode) {
          textNode.data = await minifyCSS(textNode.data, '', this.options);
        }
        break;
      }
    }
  }

  processComment(node) {
    if (this.options.removeComments) {
      removeComments$1(node);
    }
  }

  processDirective(node) {
    switch (node.name) {
      case '!doctype':
        useShortDoctype(node, this.options);
        break;
    }
  }

  async minify(tree) {
    const nodes = Array.isArray(tree) ? tree : [tree];

    for (let i = 0; i < nodes.length; i++) {
      const node = nodes[i];

      switch (node.type) {
        case htmlparser2.ElementType.Root:
          await this.minify(node.children);
          break;

        case htmlparser2.ElementType.Text:
          await this.processText(node);
          break;

        case htmlparser2.ElementType.Tag:
          await this.processTag(node);
          break;

        case htmlparser2.ElementType.Comment:
          this.processComment(node);
          break;

        case htmlparser2.ElementType.Directive:
          this.processDirective(node);
          break;
      }
    }

    return tree;
  }
}

class Sorter {
  sort(tokens, fromIndex = 0) {
    for (let i = 0, len = this.keys.length; i < len; i++) {
      const key = this.keys[i];
      const token = key.slice(1);

      let index = tokens.indexOf(token, fromIndex);

      if (index !== -1) {
        do {
          if (index !== fromIndex) {
            tokens.splice(index, 1);
            tokens.splice(fromIndex, 0, token);
          }

          fromIndex++;
        } while ((index = tokens.indexOf(token, fromIndex)) !== -1);

        return this[key].sort(tokens, fromIndex);
      }
    }

    return tokens;
  }
}

class TokenChain {
  add(tokens) {
    tokens.forEach((token) => {
      const key = '$' + token;

      if (!this[key]) {
        this[key] = [];
        this[key].processed = 0;
      }

      this[key].push(tokens);
    });
  }

  createSorter() {
    const sorter = new Sorter();

    sorter.keys = Object.keys(this)
      .sort((j, k) => {
        const m = this[j].length;
        const n = this[k].length;
        return m < n ? 1 : m > n ? -1 : j < k ? -1 : j > k ? 1 : 0;
      })
      .filter((key) => {
        if (this[key].processed < this[key].length) {
          const token = key.slice(1);
          const chain = new TokenChain();

          this[key].forEach((tokens) => {
            let index;

            while ((index = tokens.indexOf(token)) !== -1) {
              tokens.splice(index, 1);
            }

            tokens.forEach((token) => {
              this['$' + token].processed++;
            });

            chain.add(tokens.slice(0));
          });

          sorter[key] = chain.createSorter();
          return true;
        }

        return false;
      });

    return sorter;
  }
}

const attrNames = (attrs, options = defaultOptions) => {
  return attrs.map(attr => options.name(attr.name));
};

const createSortFunctions = (tree, options = defaultOptions, ctx = defaultContext, attrChains, classChain) => {
  const nodes = Array.isArray(tree) ? tree : [tree];

  const uidIgnore = ctx.get('uidIgnore');
  const uidAttr = ctx.get('uidAttr');

  const shouldSkipUID = (token, uid) => {
    return !uid || token.indexOf(uid) === -1;
  };

  const shouldSkipUIDs = (token) => {
    return shouldSkipUID(token, uidIgnore) && shouldSkipUID(token, uidAttr);
  };

  for (let i = 0; i < nodes.length; i++) {
    const node = nodes[i];

    switch (node.type) {
      case htmlparser2.ElementType.Root:
        createSortFunctions(node.children, options, ctx, attrChains, classChain);
        break;

      case htmlparser2.ElementType.Tag: {
        const { name: tag, attrs = [] } = node;

        if (attrChains) {
          if (!attrChains[tag]) {
            attrChains[tag] = new TokenChain();
          }

          attrChains[tag].add(attrNames(attrs, options).filter(shouldSkipUIDs));
        }

        attrs.forEach(attr => {
          if (classChain && attr.value && options.name(attr.name) === 'class') {
            classChain.add(trimWhitespace(attr.value).split(/[ \t\n\f\r]+/));
          }
        });

        if (node.children.length) {
          createSortFunctions(node.children, options, ctx, attrChains, classChain);
        }
        break;
      }
    }
  }
};

const applySort = (tree, options = defaultOptions, sortAttributes, sortClassName) => {
  const nodes = Array.isArray(tree) ? tree : [tree];

  for (let i = 0; i < nodes.length; i++) {
    const node = nodes[i];

    switch (node.type) {
      case htmlparser2.ElementType.Root:
        applySort(node.children, options, sortAttributes, sortClassName);
        break;

      case htmlparser2.ElementType.Tag:

        if (node.attrs && (sortAttributes || sortClassName)) {
          if (sortAttributes) {
            node.attrs = sortAttributes(node.name, node.attrs);
          }

          if (sortClassName) {
            node.attrs = node.attrs.map(attr => {
              if (options.name(attr.name) === 'class' && sortClassName) {
                return {
                  ...attr,
                  value: sortClassName(attr.value)
                };
              }

              return attr;
            });
          }
        }

        if (node.children.length) {
          applySort(node.children, options, sortAttributes, sortClassName);
        }
        break;
    }
  }
};

const createSorter = (tree, options = defaultOptions, ctx = defaultContext) => {
  const attrChains = ctx.get('attrChains') ?? (options.sortAttributes && Object.create(null));
  const classChain = ctx.get('classChain') ?? (options.sortClassName && new TokenChain());

  createSortFunctions(tree, options, ctx, attrChains, classChain);

  if (!attrChains && !classChain) {
    return;
  }

  let sortAttributesFn = null;
  if (attrChains) {
    const attrSorters = Object.create(null);

    for (const tag in attrChains) {
      attrSorters[tag] = attrChains[tag].createSorter();
    }

    const defaultAttrsSorter = (tag, attrs) => {
      const sorter = attrSorters[tag];

      if (sorter) {
        const attrMap = Object.create(null);

        const names = attrNames(attrs);
        names.forEach(function (name, index) {
          (attrMap[name] || (attrMap[name] = [])).push(attrs[index]);
        });

        sorter.sort(names).forEach(function (name, index) {
          attrs[index] = attrMap[name].shift();
        });
      }

      return attrs;
    };

    sortAttributesFn = typeof options.sortAttributes === 'function'
      ? options.sortAttributes
      : defaultAttrsSorter;
  }

  let sortClassNameFn = null;
  if (classChain) {
    const sorter = classChain.createSorter();

    const defaultClassNameSorter = (value) => {
      return sorter.sort(value.split(/[ \n\f\r]+/)).join(' ');
    };

    sortClassNameFn = typeof options.sortClassName === 'function'
      ? options.sortClassName
      : defaultClassNameSorter;
  }

  if (!sortAttributesFn && !sortClassNameFn) {
    return;
  }

  ctx.set('attrChains', attrChains);
  ctx.set('classChain', classChain);

  return () => {
    applySort(tree, options, sortAttributesFn, sortClassNameFn);
  };
};

const endsWithTrailingSlash = (str = '') => /\/$/.test(str);

const formatAttributes = (elem, options = defaultOptions) => {
  const { attrs } = elem;

  if (!attrs) {
    return;
  }

  return attrs.reduce((attrString, { name, value, quote }, index) => {
    let attr = attrString;

    attr += name;

    if (typeof quote !== 'undefined') {
      attr += `=${quote}${value}${quote}`;
    }

    const isLast = attrs.length === index + 1;

    if (isLast && quote === '' && selfClosingTags.has(elem.name)) {
      attr += SINGLE_SPACE;
    }

    if (options.removeTagWhitespace) {
      if (!isLast && (!quote || value.length === 0)) {
        attr += SINGLE_SPACE;
      }
    } else {
      if (!isLast) {
        attr += SINGLE_SPACE;
      }
    }

    // make sure trailing slash is not interpreted as HTML self-closing tag
    if (!quote && endsWithTrailingSlash(attr)) {
      attr += SINGLE_SPACE;
    }

    return attr;
  }, '');
};

class Serializer {
  constructor(options = defaultOptions) {
    this.options = options;
  }

  render(tree) {
    const nodes = Array.isArray(tree) ? tree : [tree];

    let output = '';

    for (let i = 0; i < nodes.length; i++) {
      output += this.renderNode(nodes[i]);
    }

    return output;
  }

  renderNode(node) {
    switch (node.type) {
      case htmlparser2.ElementType.Root:
        return this.render(node.children);

      case htmlparser2.ElementType.Tag:
      case htmlparser2.ElementType.Script:
        return this.renderTag(node);

      case htmlparser2.ElementType.Text:
        return this.renderText(node);

      case htmlparser2.ElementType.Comment:
        return this.renderComment(node);

      case htmlparser2.ElementType.Directive:
        return this.renderDirective(node);

      default:
        return '';
    }
  }

  renderTag(elem) {
    let openTag = `<${elem.name}`;

    const attrs = formatAttributes(elem, this.options);
    if (attrs) {
      openTag += ` ${attrs}`;
    }

    if (elem.children.length === 0 && this.options.keepClosingSlash) {
      openTag += '/>';
    } else {
      openTag += '>';
    }

    if (!this.options.includeAutoGeneratedTags && elem.startImplied) {
      openTag = '';
    }

    let children = '';
    if (openTag && elem.children) {
      children = this.render(elem.children);
    }

    let closeTag = !voidTags.has(elem.name) ? `</${elem.name}>` : '';
    if (!this.options.includeAutoGeneratedTags && elem.endImplied) {
      closeTag = '';
    }

    return openTag + children + closeTag;
  }

  renderText(elem) {
    return elem.data;
  }

  renderComment(elem) {
    return `<!--${elem.data}-->`;
  }

  renderDirective(elem) {
    return `<${elem.data}>`;
  }
}

const processScripts = async (tree, options = defaultOptions, minify) => {
  const nodes = Array.isArray(tree) ? tree : [tree];

  for (let i = 0; i < nodes.length; i++) {
    const node = nodes[i];

    switch (node.type) {
      case htmlparser2.ElementType.Root:
        await processScripts(node.children, options, minify);
        break;

      case htmlparser2.ElementType.Tag: {
        if (options.processScripts && specialContentTags.has(node.name)) {
          const attr = node.attrs?.find(attr => options.name(attr.name) === 'type');

          if (options.processScripts.indexOf(attr?.value) > -1) {
            const textNode = node.children[0];
            if (textNode) {
              textNode.data = await minify(textNode.data);
            }
          }
        }
        break;
      }
    }
  }
};

new Set([...inlineTags, ...selfClosingTags]);

const _collapseWhitespace = (str = EMPTY, node, options = defaultOptions) => {
  let text = str;

  // skip processing empty strings
  if (text.length === 0) {
    return text;
  }

  const { prev, next, parent } = node;


  let trimLeft = false;
  let trimRight = false;
  const isParentTag = parent.type === htmlparser2.ElementType.Tag;

  // strip non space whitespace then compress spaces to one
  // elements inside tags
  const collapseAll = (isParentTag || parent.type === htmlparser2.ElementType.Root) && !prev && !next;

  return collapseWhitespace(text, options, trimLeft, trimRight, collapseAll);
};

const processWhitespace = (tree, options = defaultOptions) => {
  const nodes = Array.isArray(tree) ? tree : [tree];

  for (let i = 0; i < nodes.length; i++) {
    const node = nodes[i];

    switch (node.type) {
      case htmlparser2.ElementType.Root:
        processWhitespace(node.children, options);
        break;

      case htmlparser2.ElementType.Tag:
        if (node.children.length) {
          processWhitespace(node.children, options);
        }
        break

      case htmlparser2.ElementType.Text:
        if (options.collapseWhitespace) {
          node.data = _collapseWhitespace(node.data, node, options);
        }
        break;
    }
  }
};

const parseDocument = (data, parserOptions) => {
  const handler = new DomHandler();
  new htmlparser2.Parser(handler, parserOptions).end(data);
  return handler.root;
};

const _minify = async (input = '', opts = defaultOptions, ctx = defaultContext) => {
  const options = { ...defaultOptions, ...opts };

  const context = new Context();
  context.extend(ctx);

  const { html, restore } = preProcessInput(input, options, context);

  const parserOptions = {
    recognizeSelfClosing: true,
    decodeEntities: options.decodeEntities,
    lowerCaseAttributeNames: !options.caseSensitive
  };

  const tree = parseDocument(html, parserOptions);

  const htmlMinifier = new Minifier(options, context);
  await htmlMinifier.minify(tree);

  if (options.sortAttributes || options.sortClassName) {
    const sort = createSorter(tree, options, context);
    sort();
  }

  if (options.processScripts) {
    const minifier = async (text) => {
      return _minify(text, options, context);
    };

    await processScripts(tree, options, minifier);
  }

  if (options.collapseWhitespace) {
    processWhitespace(tree, options);
  }

  const serializer = new Serializer(options);
  const output = serializer.render(tree);

  return restore(output);
};

const minify = async (input = '', options = defaultOptions) => {
  return _minify(input, options);
};

exports.minify = minify;
